;;;; Extended example adapted slightly from that given in section 9.7 of
;;;; "A Programmer's Guide to COMMON LISP" by Deborah G. Tatar,
;;;; Digital Press, 1987.  ISBN 0-932376-87-8.
;;;;
;;;; Modifications are mainly due to the lack of &rest and &optional args,

;;; Get some Common Lispy things
(load "common-lisp")

;;; some internal parameters
(defparameter *form-command-char* $%)
(defvar *input-stream*)
(defvar *output-stream*)

;;; some user-interface parameters
(defvar *current-line-number*)
(defvar *current-column*)
(defvar *page-number*)
(defvar *fill-mode?*)
(defvar *page-length*)
(defvar *left-margin*)
(defvar *text-width*)
(defvar *header-text-list*)

;;; initialize the user-interface parameters to nice defaults
(defun form-initialize ()
	(setq *current-line-number* 0)
	(setq *current-column* 0)
	(setq *page-number* 1)
	(setq *fill-mode?* f)
	(setq *page-length* 47)
	(setq *left-margin* 8)
	(setq *text-width* 70)
	(setq *header-text-list* '("")))

;;; this is the function the user calls
(defun form (input-filename output-filename)
	(form-initialize)
	(with-open-file (*output-stream* output-filename :direction :output)
		(with-open-file (*input-stream* input-filename :direction :input)
			(form-driver-loop))))

;;; main loop
(defun form-driver-loop ()
	(do	((current-line	(read-line *input-stream* 'form-driver-loop-eof)))
		;; current-line will be the symbol FORM-DRIVER-LOOP-EOF
		;; at end-of-file.
		(eq current-line 'form-driver-loop-eof)
		;; a blank line can't be a command line
		(cond	((and	(> (length current-line) 0)
						(equal	(char current-line 0)
								*form-command-char*))
						(form-execute-command current-line))
				(t	(if *fill-mode?*
						(form-fill-output-line current-line)
						(form-no-fill-output-line current-line))))
		(setq current-line (read-line *input-stream* 'form-driver-loop-eof))))

;;; output a line of input text in no-fill mode
(defun form-no-fill-output-line (line)
	(form-output-left-margin)				; first  output the left margin
	(format *output-stream* "~A" `(,line))		; then the text
	(form-newline 1))						; then a new line, and page if necessary

;;; output a line of input text in fill mode
(defun form-fill-output-line (line)
	(do	((pointer 0))
		false	; this loop only exits through return
		(multiple-value-bind (word char-pointer)
			(form-get-word pointer line)
			(if	(null word)
				(return))
			;; see if the word fits on this ine,
			;; and do a newline if it doesn't
			(if	(>= (+ *current-column* (length word)) *text-width*)
				(form-newline 1))			; output newline, check for new page
			;; output the margin spacing if at beginning of a new line
			(if	(zerop *current-column*)
				(form-output-left-margin))
			;;; now print the word
			(format *output-stream* "~A" `(,word))
			;; and the trailing space, if there's room
			(unless	(= *text-width* (+ *current-column* (length word)))
					(write-char $  *output-stream*))
			(setq *current-column* (+ *current-column* (+ (length word) 1)))
			;; the next call to form-get-word should start
			;; after the last word
			(setq pointer char-pointer))))

(defun form-get-word (pointer line)
	;; strip any leadinnnng white space
	(let	((word-beginning
			(position-if-not 'white-space-char? line pointer)))
		;; now find the end of the word
		;; i.e. the first char not in it
		(if (not (null word-beginning))			;if we found a word
			(let	((word-end (position-if 'white-space-char? line word-beginning)))
				;; if word-end's nil, it's the end of a string
				(if	(null word-end)
					(setq word-end (length line)))
				(values	(subseq line word-beginning word-end)
						word-end))
			nil)))

;;; output the left margin spacing
(defun form-output-left-margin ()
	(dotimes (i *left-margin*)
		(write-char $  *output-stream*)))

;;; output a new line and a new page (including any header)
;;; if necessary
(defun form-newline (times)
	(dotimes (i (if (null times) 1 times))
		(setq *current-column* 0)
		(incq *current-line-number*)
		;; have we reached the maximum number of lines for a page?
		;; if so, begin a new page
		(if	(= *current-line-number* *page-length*)
			(form-newpage 1)
			;; otherwise just output a new-line char
			(format *output-stream* "~%" nil))))

;;; output a new page and header, if any
(defun form-newpage (times)
	(dotimes (i (if (null times) 1 times))
		(format *output-stream* "~|" nil)
		(incq *page-number*)
		(setq *current-line-number* 0)
		(setq *current-column* 0)
		(form-output-header)))

;;; output the header text which is a list containing a
;;; format control string (which may contain FORM commands)
;;; and a set of args.  FORMAT will be called on the control
;;; string and args, and the resulting string will be treated
;;; as input text in a recursive invocation of the driver loop.
;;; Before formatting the header text, the values of
;;; "fill-mode?* and *left-margin* are saved and then restored
;;; after the header text is output.
(defun form-output-header ()
	(if t (return nil))
	(let	((header-string	"THIS ISN'T IMPLEMENTED YET")		;(apply format (mapcar eval *header-text-list*)))
		 (fill-mode?		*fill-mode?*)	
		 (left-margin		*left-margin*)
		 (input-stream	*input-stream*))
		(setq *input-stream* (input-string-stream header-string))
		(unwind-protect (form-driver-loop)
			(setq *fill-mode?* fill-mode?)		;shouldn't this be done using fluid-let??!!!!
			(setq *left-margin* left-margin)
			(setq *input-stream* input-stream))))

;;; form-execute-command executes a user command
(defun form-execute-command (current-line)
	(let*		((command-list	(list-from-string (subseq current-line 1 nil)))
			 (command		(car command-list))
			 (args			(cdr command-list)))
		(case command
			;; it would be better to make this data driven
			;; but for a small number of commands, CASE will do.
			(fill				(setq *fill-mode?* t))
			(no-fill			(form-newline 1) (setq *fill-mode?* f))
			(left-margin		(setq *left-margin* (car args)))
			(new-page		(apply form-newpage args))
			;(header			(setq *header-text-list* args))
			(text-width		(setq *text-width* (car args)))
			(page-length	(setq *page-length* (car args)))
			(break			(apply form-newline args))
			(eval			(mapc 'eval args))
			(t				(format t "~%Warning: ~S is not a known FORM command." `(,command))))))

;;; some utility functions

(defun white-space-char? (char)
	(or	(char-equal char $ )
		(char-equal char $	)))

;;; LIST-FROM-STRING creates a list of the tokens read from a
;;; string and returns it

;; we gonna do this as a prim'tive fer now - I'm bone idle!
(primitive (list-from-string _x)
	"^(LispParser new
		parse: (ReadStream on: '(' , _x , ')')
		editor: nil) first")
;;
;; and here's what it aughtta be...
;;(defun list-from-string (string)
;;	(let	((input	(input-string-stream string)))
;;		(do	((token	(read input nil 'list-from-string-eof)
;;					(read input nil 'list-from-string eof))
;;			 (list))
;;			((eq token 'list-from-string-eof) (nreverse list))
;;			(push token list))))

;;; That's it - all that remains is a bit of smirkery

(load "interface")		; gets the prompters/confirmers/etc...

(defun run-form ()
	(let	((in		(promptForWithDefault "input filename?" "examples/example.text"))
		 (out	(promptForWithDefault "output filename?" "examples/formatted.text")))
		(form in out)
		(if (promptForYesNo "Do you want to inspect the output?")
			(inspect-file-contents out))))

(if (promptForYesNo
"  Call the function ""(run-form)"" to run the formatter.
    You will be promted for a filename - a suitable
default will be supplied.  Do you want to run it now?")
	(run-form))
